@@ -129,11 +129,12 @@ public abstract class URLStreamHandler
     
     if (spec.regionMatches (start, "//", 0, 2))
       {
+	String genuineHost;
 	int hostEnd;
-	int colon;
+	int colon, at_host;
 
 	start += 2;
-	int slash = spec.indexOf('/', start);
+	int slash = spec.indexOf ('/', start);
 	if (slash >= 0) 
 	  hostEnd = slash;
         else
@@ -141,24 +142,37 @@ public abstract class URLStreamHandler
 
 	host = spec.substring (start, hostEnd);
 	
+	// We first need a genuine host name (with userinfo).
+	// So we check for '@': if it's present check the port in the
+	// section after '@' in the other case check it in the full string.
+	// P.S.: We don't care having '@' at the beginning of the string.
+	if ((at_host = host.indexOf ('@')) >= 0)
+	  genuineHost = host.substring (at_host);
+	else
+	  genuineHost = host;
+
 	// Look for optional port number.  It is valid for the non-port
 	// part of the host name to be null (e.g. a URL "http://:80").
 	// TBD: JDK 1.2 in this case sets host to null rather than "";
 	// this is undocumented and likely an unintended side effect in 1.2
 	// so we'll be simple here and stick with "". Note that
 	// "http://" or "http:///" produce a "" host in JDK 1.2.
-	if ((colon = host.indexOf(':')) >= 0)
+	if ((colon = genuineHost.indexOf (':')) >= 0)
 	  {
 	    try
 	      {
-		port = Integer.parseInt(host.substring(colon + 1));
+		port = Integer.parseInt (genuineHost.substring (colon + 1));
 	      }
 	    catch (NumberFormatException e)
 	      {
 		; // Ignore invalid port values; port is already set to u's
 		  // port.
 	      }
-	    host = host.substring(0, colon);
+	    // Now we must cut the port number in the original string.
+	    if (at_host >= 0)
+	      host = host.substring (0, at_host + colon);
+	    else
+	      host = host.substring (0, colon);
 	  }
 	file = null;
 	start = hostEnd;
@@ -176,7 +190,7 @@ public abstract class URLStreamHandler
       } 
     else if (start < end)
       {
-	// Context is available, but only override it if there is a new file.
+        // Context is available, but only override it if there is a new file.
         char sepChar = '/';
         int lastSlash = file.lastIndexOf (sepChar);
         if (lastSlash < 0 && File.separatorChar != sepChar
@@ -196,10 +210,15 @@ public abstract class URLStreamHandler
             // need to canonicalise the file path.
             try
               {
+		boolean endsWithSlash = file.charAt(file.length() - 1) == '/';
                 file = new File (file).getCanonicalPath ();
+		if (endsWithSlash
+		    && file.charAt(file.length() - 1) != '/')
+		  file += '/';
               }
             catch (IOException e)
               {
+		// Do nothing.
               }
           }
 
@@ -225,6 +244,9 @@ public abstract class URLStreamHandler
     setURL(url, url.getProtocol(), host, port, file, ref);
   }
   
+  /*
+   * Canonicalize a filename.
+   */
   private static String canonicalizeFilename(String file)
   {
     // XXX - GNU Classpath has an implementation that might be more appropriate
@@ -257,6 +279,8 @@ public abstract class URLStreamHandler
    * @param url1 The first url
    * @param url2 The second url to compare with the first
    * 
+   * @return True if both URLs point to the same file, false otherwise.
+   *
    * @specnote Now protected
    */
   protected boolean sameFile(URL url1, URL url2)
@@ -265,7 +289,15 @@ public abstract class URLStreamHandler
       return true;
     // This comparison is very conservative.  It assumes that any
     // field can be null.
-    if (url1 == null || url2 == null || url1.getPort() != url2.getPort())
+    if (url1 == null || url2 == null)
+      return false;
+    int p1 = url1.getPort ();
+    if (p1 == -1)
+      p1 = url1.ph.getDefaultPort ();
+    int p2 = url2.getPort ();
+    if (p2 == -1)
+      p2 = url2.ph.getDefaultPort ();
+    if (p1 != p2)
       return false;
     String s1, s2;
     s1 = url1.getProtocol();
@@ -337,6 +369,8 @@ public abstract class URLStreamHandler
    *
    * @param url1 An URL object
    * @param url2 An URL object
+   *
+   * @return True if both given URLs are equal, false otherwise.
    */
   protected boolean equals (URL url1, URL url2)
   {
@@ -369,26 +403,43 @@ public abstract class URLStreamHandler
   /**
    * Compares the host components of two URLs.
    *
+   * @param url1 The first URL.
+   * @param url2 The second URL.
+   *
+   * @return True if both URLs contain the same host.
+   *
    * @exception UnknownHostException If an unknown host is found
    */
   protected boolean hostsEqual (URL url1, URL url2)
-    throws UnknownHostException
   {
-    InetAddress addr1 = InetAddress.getByName (url1.getHost ());
-    InetAddress addr2 = InetAddress.getByName (url2.getHost ());
+    InetAddress addr1 = getHostAddress (url1);
+    InetAddress addr2 = getHostAddress (url2);
 
-    return addr1.equals (addr2);
+    if (addr1 != null || addr2 != null)
+      return addr1.equals (addr2);
+
+    String host1 = url1.getHost();
+    String host2 = url2.getHost();
+    
+    if (host1 != null && host2 != null)
+      return host1.equalsIgnoreCase (host2);
+
+    return host1 == null && host2 == null;
   }
 
   /**
    * Get the IP address of our host. An empty host field or a DNS failure will
    * result in a null return.
+   *
+   * @param url The URL to return the host address for.
+   *
+   * @return The address of the hostname in url.
    */
   protected InetAddress getHostAddress (URL url)
   {
     String hostname = url.getHost ();
 
-    if (hostname == "")
+    if (hostname.equals(""))
       return null;
     
     try
@@ -404,6 +455,8 @@ public abstract class URLStreamHandler
   /**
    * Returns the default port for a URL parsed by this handler. This method is
    * meant to be overidden by handlers with default port numbers.
+   *
+   * @return The default port number.
    */
   protected int getDefaultPort ()
   {
@@ -413,6 +466,10 @@ public abstract class URLStreamHandler
   /**
    * Provides the default hash calculation. May be overidden by handlers for
    * other protocols that have different requirements for hashCode calculation.
+   *
+   * @param url The URL to calc the hashcode for.
+   * 
+   * @return The hashcode for the given URL.
    */
   protected int hashCode (URL url)
   {
@@ -428,10 +485,12 @@ public abstract class URLStreamHandler
    * that have a different syntax should override this method
    *
    * @param url The URL object to convert
+   *
+   * @return A string representation of the url
    */
   protected String toExternalForm(URL u)
   {
-    String protocol, host, file, ref;
+    String protocol, host, file, ref, user;
     int port;
 
     protocol = u.getProtocol();
@@ -445,25 +504,30 @@ public abstract class URLStreamHandler
     port = u.getPort();
     file = u.getFile();
     ref = u.getRef();
+    user = u.getUserInfo();
 
     // Guess a reasonable size for the string buffer so we have to resize
     // at most once.
     int size = protocol.length() + host.length() + file.length() + 24;
     StringBuffer sb = new StringBuffer(size);
 
-    sb.append(protocol);
-    sb.append(':');
+    if (protocol != null && protocol.length() > 0)
+      {
+	sb.append(protocol);
+	sb.append(":");
+      }
 
     if (host.length() != 0)
-      sb.append("//").append(host);
+      {
+	sb.append("//");
+	if (user != null && !"".equals(user))
+	  sb.append(user).append('@');
+	sb.append(host);
 
-    // Note that this produces different results from JDK 1.2 as JDK 1.2
-    // ignores a non-default port if host is null or "".  That is inconsistent
-    // with the spec since the result of this method is spec'ed so it can be
-    // used to construct a new URL that is equivalent to the original.
-    boolean port_needed = port > 0 && port != getDefaultPort();
-    if (port_needed)
-      sb.append(':').append(port);
+        // Append port if port was in URL spec.
+        if (port >= 0)
+          sb.append(':').append(port);
+      }
 
     sb.append(file);
 
